import numpy as np
from scipy.interpolate import interp1d
from blast.interfaces.DDataStructure import Group
import time

class SolversInterface:
    def __init__(self, _vSolver, _cfg):
        self.vSolver = _vSolver
        self.cfg = _cfg

        if 'nSections' not in self.cfg:
            self.cfg['nSections'] = len(self.cfg['sections'])
        
        if 'sweep' not in self.cfg:
            self.cfg['sweep'] = False

        # Initialize data structures.
        self.iBnd = [Group('iWing' if iReg == 0 else 'iWake') for iReg in range(2)]
        self.vBnd = [[Group('vAirfoil'+str(k) if iReg == 0 else 'vWake'+str(k)) for iReg in range(2)] for k in range(self.cfg['nSections'])]
        
        if self.cfg['interpolator'] == 'Matching':
            from blast.interfaces.interpolators.DMatchingInterpolator import MatchingInterpolator as interp
        elif self.cfg['interpolator'] == 'Rbf':
            from blast.interfaces.interpolators.DRbfInterpolator import RbfInterpolator as interp
            print('RBF interpolator initialized.')
            print('Loading viscous mesh... ', end='')
            initialTime = time.time()
            if self.cfg['nDim'] == 2:
                self.getVAirfoil()
            elif self.cfg['nDim'] == 3:
                self.getVWing()
            print('done in {:.2f}s.'.format(time.time()-initialTime))
        else:
            raise RuntimeError('Incorrect interpolator specified in config.')
        self.interpolator = interp(self.cfg)

        print('Loading inviscid mesh... ', end='')
        initialTime = time.time()
        self.setMesh()
        print('done in {:.2f}s.'.format(time.time()-initialTime))

        #self.interpolator.ilwIdx = self.ilwIdx
        #self.interpolator.vlwIdx = self.vlwIdx

        self.deltaStarExt = [[np.zeros(0) for iReg in range(3)]\
                             for iSec in range(self.cfg['nSections'])]
        self.xxExt        = [[np.zeros(0) for iReg in range(3)]\
                             for iSec in range(self.cfg['nSections'])]
        #self.vtOld = [[np.zeros(0) for iReg in range(3)] for iSec in range(self.cfg['nSections'])]
    
    def setViscousSolver(self, couplIter):
        """Interpolate solution to viscous mesh and sets the inviscid boundary in the viscous solver.
        """
        self.interpolator.inviscidToViscous(self.iBnd, self.vBnd, couplIter)
        for iSec in range(self.cfg['nSections']):
            ## Mesh
            if (self.vBnd[iSec][0].isMeshChanged()):
                for reg in self.vBnd[iSec]:
                    if 'vWake' in reg.name:
                        iReg = 2
                        self.vSolver.setGlobMesh(iSec, iReg,
                                                 reg.nodesCoord[:,0],
                                                 reg.nodesCoord[:,1],
                                                 reg.nodesCoord[:,2])
                        # External variables
                        self.xxExt[iSec][2] = np.zeros(reg.nPoints)
                        self.deltaStarExt[iSec][2] = np.zeros(reg.nPoints)
                        #self.vtOld[iSec][2] = np.zeros(reg.nPoints)
                    elif 'vAirfoil' in reg.name:
                        self.vSolver.setGlobMesh(iSec, 0,
                                                 reg.getXUpper(),
                                                 reg.getYUpper(),
                                                 reg.getZUpper())
                        self.vSolver.setGlobMesh(iSec, 1,
                                                 abs(reg.getXLower()),
                                                 reg.getYLower(),
                                                 reg.getZLower())
                        # External variables
                        self.xxExt[iSec][0]        = np.zeros(reg.stagPoint+1)
                        self.deltaStarExt[iSec][0] = np.zeros(reg.stagPoint+1)
                        #self.vtOld[iSec][0]        = np.zeros(reg.stagPoint+1)
                        self.xxExt[iSec][1]        = np.zeros(reg.nPoints
                                                              - reg.stagPoint)
                        self.deltaStarExt[iSec][1] = np.zeros(reg.nPoints
                                                              - reg.stagPoint)
                        #self.vtOld[iSec][1]        = np.zeros(reg.nPoints
                        #                                     - reg.stagPoint)
                    else:
                        raise RuntimeError('Tried to initialized viscous\
                                           struture but name was', reg.name)
            ## Inviscid variables
            for reg in self.vBnd[iSec]:
                reg.updatePrev()
                if 'vWake' in reg.name:
                    iReg = 2
                    self.vSolver.setVelocities(iSec, iReg,
                                               reg.V[:,0],
                                               reg.V[:,1],
                                               reg.V[:,2])
                    self.vSolver.setMe(iSec, iReg, reg.M)
                    self.vSolver.setRhoe(iSec, iReg, reg.Rho)
                elif 'vAirfoil' in reg.name:
                    iReg = 0 # Upper side
                    self.vSolver.setVelocities(iSec, iReg,
                                               reg.getVelocityXUpper(),
                                               reg.getVelocityYUpper(),
                                               reg.getVelocityZUpper())
                    self.vSolver.setMe(iSec, iReg, reg.getMachUpper())
                    self.vSolver.setRhoe(iSec, iReg, reg.getRhoUpper())
                    iReg = 1 # Lower side
                    self.vSolver.setVelocities(iSec, iReg,
                                               reg.getVelocityXLower(),
                                               reg.getVelocityYLower(),
                                               reg.getVelocityZLower())
                    self.vSolver.setMe(iSec, iReg, reg.getMachLower())
                    self.vSolver.setRhoe(iSec, iReg, reg.getRhoLower())
                else:
                    raise RuntimeError('Tried to initialize a viscous\
                                       struture but name was', reg.name)
            ## External variables
            for iRegv in range(3):
                self.vSolver.setxxExt(iSec, iRegv, self.xxExt[iSec][iRegv])
                self.vSolver.setDeltaStarExt(iSec, iRegv,
                                             self.deltaStarExt[iSec][iRegv])
                #self.vSolver.setVtOld(iSec, iRegv, self.vtOld[iSec][iRegv])
    def getBlowingBoundary(self):
        for iSec in range(len(self.vBnd)):
            for reg in self.vBnd[iSec]:
                if 'vWake' in reg.name:
                    reg.blowingVel = np.asarray(self.vSolver.extractBlowingVel(iSec, 2))
                elif 'vAirfoil' in reg.name:
                    reg.blowingVel = np.concatenate((np.flip(np.asarray(self.vSolver.extractBlowingVel(iSec, 0))),
                                                     np.asarray(self.vSolver.extractBlowingVel(iSec, 1))))
            for iRegv in range(3):
                self.xxExt[iSec][iRegv] = self.vSolver.extractxx(iSec, iRegv)
                self.deltaStarExt[iSec][iRegv] = self.vSolver.extractDeltaStar(iSec, iRegv)
                #self.vtOld[iSec][iRegv] = self.vSolver.extractUe(iSec, iRegv)
                
        self.interpolator.viscousToInviscid(self.iBnd, self.vBnd)

    def getVWing(self):
        """Obtain the nodes' location and row and cg of all elements.
        If the mesh is the viscous one, TE nodes are duplicated (by creating a new row).
        They exist for the interpolation and for the solution computation.
        If not, the nodes are pseudo-duplicated (without changing the row). They therefore
        exist only for the interpolation but not for the solution.
        """

        data = [np.zeros((0,4)) for _ in range(2)]

        leNodes = np.zeros((0, 3))
        lwRows = []
        ctlw = 0

        for e in self.cfg['vMsh'].elems:
            if e.ptag.name == 'wing' or e.ptag.name == 'wing_'\
                or e.ptag.name == 'leadingEdge':
                iRegion = 0
            elif e.ptag.name == 'wake' or e.ptag.name == 'wake_':
                iRegion = 1
            else:
                continue
            
            # Get nodes
            for nw in e.nodes:
                if nw.row not in data[iRegion][:,3]:
                    data[iRegion] = np.vstack((data[iRegion], [nw.pos[0],\
                                                                nw.pos[2],\
                                                                nw.pos[1],\
                                                                nw.row]))
                    # If wing_ exists, the lower side is identifiable.
                    if e.ptag.name == 'wing_':
                        lwRows.append(nw.row)
        
        if len(lwRows) == 0:
            raise RuntimeError('Could not identify lower side.')

        self.vlwIdx = []
        for idx in range(len(data[0][:,3])):
            if data[0][idx, 3] in lwRows:
                self.vlwIdx.append(idx)

        maxRowNb = max(data[0][:,3])
        # Initialize viscous strutures
        self.cfg['EffSections'] = np.empty(0)

        for iReg, reg in enumerate(data):
            for iy, y in enumerate(self.cfg['sections']):
                # Find nearest value in the mesh
                try:
                    idx = (np.abs(reg[:,2]-y).argmin())
                except:
                    print('Region', reg[:,2], 'cannot be found through\
                          config input', y)
                    raise RuntimeError("Could not identify section.\
                                       Possibly a missing tag")
                section = reg[idx,2]
                if iReg == 0:
                    self.cfg['EffSections'] = np.append(self.cfg['EffSections'], section)
                nodesSec = reg[abs(reg[:,2]-section)<1e-3, :]

                # Separate points on upper and lower side
                if iReg == 0:
                    lower_points = []
                    upper_points = []
                    for point in nodesSec:
                        if point[3] in lwRows:
                            lower_points.append(point)
                        else:
                            upper_points.append(point)

                    # Sort points in selig format
                    lower_points = sorted(lower_points, key=lambda p: p[0])
                    upper_points = sorted(upper_points, key=lambda p: p[0], reverse=True)
                    nodesSec = np.row_stack((upper_points, lower_points))
                    nodesSec = np.row_stack((nodesSec, [nodesSec[0,0], nodesSec[0,1], nodesSec[0,2], nodesSec[0,3]+maxRowNb]))
                else:
                    nodesSec = nodesSec[nodesSec[:,0].argsort()]

                # Compute cg pos on the section (!= cg of 2D elems on the surface)
                cgSec = np.zeros((len(nodesSec[:,0])-1, 4))
                for i in range(len(cgSec)):
                    cgSec[i,:3] = (nodesSec[i,:3] + nodesSec[i+1,:3])/2
                    cgSec[i, 3] = i

                self.vBnd[iy][iReg].initStructures(nodesSec.shape[0])
                self.vBnd[iy][iReg].nodesCoord = nodesSec
                self.vBnd[iy][iReg].elemsCoord = cgSec
