#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Dart interface
# Paul Dechamps

import numpy as np

from fwk.coloring import ccolors
from blast.interfaces.DSolversInterface import SolversInterface

class DartInterface(SolversInterface):
    def __init__(self, dartCfg, vSolver, _cfg):
        try:
            from modules.dartflo.dart.api.core import initDart
            argOut = initDart(dartCfg, viscous=True)
            self.solver = argOut.get('sol') # Solver
            self.wrt = argOut.get('wrt') # Writer
            self.blw = [argOut.get('blwb'), argOut.get('blww')]
            print('Dart successfully loaded.')
        except:
            print(ccolors.ANSI_RED, 'Could not load DART. Make sure it is installed.', ccolors.ANSI_RESET)
            raise RuntimeError('Import error')

        if 'iMsh' not in _cfg:
            _cfg['iMsh'] = self.blw
        SolversInterface.__init__(self, vSolver, _cfg)
    
    def run(self):
        return self.solver.run()
    
    def writeCp(self, sfx=''):
        """ Write Cp distribution around the airfoil on file.
        """
        # Extract Cp on elements
        vElems = self.blw[0].tag.elems
        vData = self.solver.Cp
        if isinstance(vData[0], float):
            size = 1
        elif vData[0].size() == 3:
            size = 3
        else:
            raise RuntimeError('unrecognized format for vData!')
        # store data (not efficient, but OK since only meant for small 2D cases)
        data = np.zeros((vElems.size()+1,3+size))
        i = 0
        while i < vElems.size()+1:
            data[i,0] = vElems[i%vElems.size()].nodes[0].pos[0]
            data[i,1] = vElems[i%vElems.size()].nodes[0].pos[1]
            data[i,2] = vElems[i%vElems.size()].nodes[0].pos[2]
            if size == 1:
                data[i,3] = vData[vElems[i%vElems.size()].nodes[0].row]
            else:
                for j in range(size):
                    data[i,3+j] = vData[vElems[i%vElems.size()].nodes[0].row][j]
            i += 1
        print('writing data file ' + 'Cp' +sfx + '.dat')
        np.savetxt('Cp'+sfx+'.dat', data, fmt='%1.6e', delimiter=', ', header='x, y, z, Cp', comments='')
    
    def save(self, sfx='inviscid'):
        self.solver.save(self.wrt, sfx)

    def getAoA(self):
        return self.solver.pbl.alpha
    
    def getMinf(self):
        return self.solver.pbl.M_inf
    
    def getCl(self):
        return self.solver.Cl

    def getCd(self):
        return self.solver.Cd

    def getCm(self):
        return self.solver.Cm
    
    def getVerbose(self):
        return self.solver.verbose
    
    def reset(self):
        self.solver.reset()
    
    def imposeInvBC(self, couplIter):
        """ Extract the inviscid paramaters at the edge of the boundary layer
        and use it as a boundary condition in the viscous solver.

        Params
        ------
        - couplIter (int): Coupling iteration.
        """
        # Retreive parameters.
        for n in range(2):
            for iRow, row in enumerate(self.iBnd[n].nodesCoord[:,3]):
                row=int(row)
                for iDim in range(3):
                    self.iBnd[n].V[iRow,iDim] = self.solver.U[row][iDim]
                self.iBnd[n].M[iRow] = self.solver.M[row]
                self.iBnd[n].Rho[iRow] = self.solver.rho[row]
            if self.cfg['nDim']==3:
                self.iBnd[n].V[:,[1,2]] = self.iBnd[n].V[:,[2,1]]
        self.setViscousSolver(couplIter)

    def imposeBlwVel(self):
        """ Extract the solution of the viscous calculation (blowing velocity)
        and use it as a boundary condition in the inviscid solver.
        """
        self.getBlowingBoundary()
        # Impose blowing velocities.
        for n in range(len(self.iBnd)):
            if self.cfg['nDim'] == 2:
                self.iBnd[n].connectBlowingVel()
            for i, blw in enumerate(self.iBnd[n].blowingVel):
                self.blw[n].setU(i, blw)

    def setMesh(self):

        if self.cfg['nDim'] == 2:
            self.getAirfoil()
        elif self.cfg['nDim'] == 3:
            self.getWing()
        else:
            raise RuntimeError('dartInterface::Could not resolve how to set\
            the mesh. Either ''nDim'' is incorrect or ''msh'' was not set for\
            3D computations')

    def getAirfoil(self):
        """Create data structure for information transfer.
        """
        self.iBnd[0].initStructures(self.blw[0].nodes.size())
        # Node number
        N1 = np.zeros(self.blw[0].nodes.size(), dtype=int)
        # Index in boundary.nodes
        connectListNodes = np.zeros(self.blw[0].nodes.size(), dtype=int)
        # Index in boundary.elems
        connectListElems = np.zeros((self.blw[0].nodes.size()-1), dtype=int)
        data = np.zeros((self.blw[0].nodes.size(), 5))
        i = 0
        for n in self.blw[0].nodes:
            data[i,0] = n.no
            data[i,1] = n.pos[0]
            data[i,2] = n.pos[1]
            data[i,3] = n.pos[2]
            data[i,4] = n.row
            i += 1
        # Table containing the element and its nodes
        eData = np.zeros(((self.blw[0].nodes.size()-1),3), dtype=int)
        elemData = np.zeros(((self.blw[0].nodes.size()-1),4))
        for i in range(0, (self.blw[0].nodes.size()-1)):
            eData[i,0] = self.blw[0].tag.elems[i].no
            eData[i,1] = self.blw[0].tag.elems[i].nodes[0].no
            eData[i,2] = self.blw[0].tag.elems[i].nodes[1].no
            elemData[i,0] = self.blw[0].tag.elems[i].nodes[0].pos[0] \
            + self.blw[0].tag.elems[i].nodes[1].pos[0]
            elemData[i,1] = self.blw[0].tag.elems[i].nodes[0].pos[1] \
            + self.blw[0].tag.elems[i].nodes[1].pos[1]
            elemData[i,2] = self.blw[0].tag.elems[i].nodes[0].pos[2] \
            + self.blw[0].tag.elems[i].nodes[1].pos[2]
            elemData[i,3] = self.blw[0].tag.elems[i].no
        elemData[i, :3] /= 2

        # Find the stagnation point
        idxStag = np.argmin(data[:,1])
        globStag = int(data[idxStag,0]) # Position of the stagnation point

        # Find TE nodes.
        # Assuming suction side element is above pressure side element in the
        # y direction.
        # TE nodes
        idxTE = np.where(data[:,1] == np.max(data[:,1]))[0]
        # TE element 1
        idxTEe0 = np.where(np.logical_or(eData[:,1] == data[idxTE[0],0], \
        eData[:,2] == data[idxTE[0],0]))[0]
        # TE element 2
        idxTEe1 = np.where(np.logical_or(eData[:,1] == data[idxTE[1],0], \
        eData[:,2] == data[idxTE[1],0]))[0]
        # y coordinates of TE element 1 CG
        ye0 = 0.5 * (data[np.where(data[:,0] == eData[idxTEe0[0],1])[0],2] \
        + data[np.where(data[:,0] == eData[idxTEe0[0],2])[0],2])
        # y coordinates of TE element 2 CG
        ye1 = 0.5 * (data[np.where(data[:,0] == eData[idxTEe1[0],1])[0],2] \
        + data[np.where(data[:,0] == eData[idxTEe1[0],2])[0],2])
        if ye0 - ye1 > 0: # element 1 containing TE node 1 is above element 2
            upperGlobTE = int(data[idxTE[0],0])
            lowerGlobTE = int(data[idxTE[1],0])
        else:
             upperGlobTE = int(data[idxTE[1],0])
             lowerGlobTE = int(data[idxTE[0],0])
        # Connectivity
        connectListElems[0] = np.where(eData[:,2] == upperGlobTE)[0]
        # number of the stag node elems.nodes -> globStag
        N1[0] = eData[connectListElems[0],1]
        #connectListNodes[0] = np.where(data[:,0] == N1[0])[0]
        connectListNodes[0] = np.where(data[:,0] == upperGlobTE)[0]
        i = 1
        upperStag = 0
        lowerTE = 0
        # Sort the suction part
        while upperStag == 0:
            N1[i] = eData[connectListElems[i-1],1] # Second node of the element
            # Index of the first node of the next element in elems.nodes
            connectListElems[i] = np.where(eData[:,2] == N1[i])[0]
            # Index of the node in boundary.nodes
            connectListNodes[i] = np.where(data[:,0] == N1[i])[0]
            if eData[connectListElems[i],1] == globStag:
                upperStag = 1
            i += 1
        self.idxRt = i+1
        # Sort the pressure side
        connectListElems[i] = np.where(eData[:,2] == globStag)[0]
        connectListNodes[i] = np.where(data[:,0] == globStag)[0]
        N1[i] = eData[connectListElems[i],2]
        while lowerTE == 0:
            # First node of the element
            N1[i+1]  = eData[connectListElems[i],1]
            # Index of the second node of the next element in elems.nodes
            connectListElems[i+1] = np.where(eData[:,2] == N1[i+1])[0]
            # Index of the node in boundary.nodes
            connectListNodes[i+1] = np.where(data[:,0] == N1[i+1])[0]
            if eData[connectListElems[i+1],1] == lowerGlobTE:
                lowerTE = 1
            i += 1
        connectListNodes[i+1] = np.where(data[:,0] == lowerGlobTE)[0]
        data[:,0] = data[connectListNodes,0]
        data[:,1] = data[connectListNodes,1]
        data[:,2] = data[connectListNodes,2]
        data[:,3] = data[connectListNodes,3]
        data[:,4] = data[connectListNodes,4]
        self.iBnd[0].nodesCoord = np.column_stack((data[:,1], data[:,2],\
                                                   data[:,3], data[:,4]))
        self.iBnd[0].setConnectList(connectListNodes, connectListElems)

        # Wake
        self.iBnd[1].initStructures(self.blw[1].nodes.size())
        # Node number
        N1 = np.zeros(self.blw[1].nodes.size(), dtype=int)
        # Index in boundary.nodes
        connectListNodes = np.zeros(self.blw[1].nodes.size(), dtype=int)
        # Index in boundary.elems
        connectListElems = np.zeros((self.blw[1].nodes.size()-1), dtype=int)
        dataW = np.zeros((self.blw[1].nodes.size(), 5))
        i = 0
        for n in self.blw[1].nodes:
            dataW[i,0] = n.no
            dataW[i,1] = n.pos[0]
            dataW[i,2] = n.pos[1]
            dataW[i,3] = n.pos[2]
            dataW[i,4] = n.row
            i += 1
        # Table containing the element and its nodes
        eData = np.zeros(((self.blw[1].nodes.size()-1),3), dtype=int)
        for i in range(0, (self.blw[1].nodes.size()-1)):
            eData[i,0] = self.blw[1].tag.elems[i].no
            eData[i,1] = self.blw[1].tag.elems[i].nodes[0].no
            eData[i,2] = self.blw[1].tag.elems[i].nodes[1].no
        connectListNodes = dataW[:,1].argsort()
        # connectListElems[0] = np.where(eData[:,1] == min(data[:,1]))[0]
        connectListElems[0] = 0
        for i in range(1, len(eData[:,0])):
            connectListElems[i] = \
            np.where(eData[:,1] == eData[connectListElems[i-1],2])[0]
        dataW[:,0] = dataW[connectListNodes,0]
        dataW[:,1] = dataW[connectListNodes,1]
        dataW[:,2] = dataW[connectListNodes,2]
        dataW[:,3] = dataW[connectListNodes,3]
        dataW[:,4] = dataW[connectListNodes,4]
        self.iBnd[1].nodesCoord = np.column_stack((dataW[:,1], dataW[:,2], \
                                                   dataW[:,3], dataW[:,4]))
        self.iBnd[1].setConnectList(connectListNodes, connectListElems)

    def getWing(self):
        """Obtain the nodes' location and row and cg of all elements.
        """

        data = [np.zeros((0,4)) for _ in range(len(self.cfg['iMsh']))]
        cg = [np.zeros((0,4)) for _ in range(len(self.cfg['iMsh']))]

        """lwRows = []
        for e in self.cfg['iMsh2'].elems:
            if e.ptag.name == 'wing_':
                for n in e.nodes:
                    if n.row not in lwRows:
                        lwRows.append(n.row)

        if len(lwRows) == 0:
            raise RuntimeError('Could not identify the lower side in the inviscid mesh.')
        """

        for iRegion in range(len(self.cfg['iMsh'])):
            for e in self.cfg['iMsh'][iRegion].tag.elems:
                # Compute cg position
                cg_pt = np.zeros(3)
                for nw in e.nodes:
                    cg_pt += [nw.pos[0], nw.pos[2], nw.pos[1]]
                    if nw.row not in data[iRegion][:,3]:
                        data[iRegion] = np.vstack((data[iRegion], [nw.pos[0],\
                                                                   nw.pos[2],\
                                                                   nw.pos[1],\
                                                                   nw.row]))
                cg_pt/=e.nodes.size()
                cg_pt = np.hstack((cg_pt, e.no))
                cg[iRegion] = np.vstack((cg[iRegion], cg_pt))

        self.ilwIdx = []
        if 'Sym' in self.cfg:
            for s in self.cfg['Sym']:
                for iReg in range(len(data)):
                    dummy = data[iReg].copy()
                    dummy[:,2] -= s
                    dummy[:,2] *= -1
                    dummy[:,2] += s
                    data[iReg] = np.row_stack((data[iReg], dummy))

        for n in range(len(data)):
            self.iBnd[n].initStructures(data[n].shape[0])
            self.iBnd[n].nodesCoord = data[n]
            self.iBnd[n].elemsCoord = cg[n]
            self.iBnd[n].setConnectList(data[n][:,3], cg[n][:,3])
