#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# SU2 interface
# Paul Dechamps
import numpy as np

from blast.interfaces.DSolversInterface import SolversInterface

class SU2Interface(SolversInterface):
    def __init__(self, fileName, vSolver, cfg):
        self.have_mpi, self.comm, self.status, self.myid = self.__setup_mpi()
        try:
            import pysu2
            self.solver = pysu2.CSinglezoneDriver(fileName, 1, self.comm)
            print('SU2 successfully loaded.')
        except:
            print(ccolors.ANSI_RED, 'Could not load SU2. Make sure it is installed.', ccolors.ANSI_RESET)
            raise RuntimeError('Import error')
        SolversInterface.__init__(self, vSolver, cfg)

    def __setup_mpi(self):
        import sys
        from mpi4py import MPI
        if 'mpi4py.MPI' in sys.modules:
            have_mpi = True
            comm = MPI.COMM_WORLD
            status = MPI.Status()
            myid = comm.Get_rank()
        else:
            have_mpi = False
            comm = None
            status = None
            myid = 0
        return have_mpi, comm, status, myid
    
    def run(self):
        self.comm.barrier()
        # run solver
        self.solver.Run()
        self.solver.Postprocess()
        # write outputs
        self.solver.Monitor(0) 
        self.solver.Output(0)
        self.comm.barrier()
        return 1

    def writeCp(self, sfx=''):
        cp = np.zeros((self.nPoints[0], 4))
        for i in range(len(self.iBnd[0].nodesCoord[:,0])):
            cp[i, :3] = self.solver.GetVertexCoord(int(self.bndMarkers[0]), int(self.iBnd[0].connectListNodes[i]))
            cp[i, 3] = self.solver.GetVertexCp(int(self.bndMarkers[0]), int(self.iBnd[0].connectListNodes[i]))
        np.savetxt('Cp'+sfx+'.dat', cp, fmt='%1.8e', delimiter=', ',
                   header='{0:>14s}, {1:>14s}, {2:>14s}, {3:>14s}'.format('x','y', 'z', 'Cp'), comments='')
    
    def save(self, writer):
        pass

    def getAoA(self):
        return 1
    
    def getMinf(self):
        return 1
    
    def getCl(self):
        return self.solver.Get_LiftCoeff()
    
    def getCd(self):
        return self.solver.Get_DragCoeff()

    def getCm(self):
        return self.solver.Get_Mx()
    
    def getVerbose(self):
        return 1
    
    def reset(self):
        self.solver.Preprocess(0)
    
    def imposeInvBC(self, couplIter):
        """ Extract the inviscid paramaters at the edge of the boundary layer
        and use it as a boundary condition in the viscous solver.

        Params
        ------
        - couplIter (int): Coupling iteration.
        """
        # Retreive parameters.
        for n in range(len(self.iBnd)):
            for iRow, row in enumerate(self.iBnd[n].connectListNodes):
                row=int(row)
                self.iBnd[n].V[iRow,:] = self.solver.GetVertexVelocity(self.bndMarkers[n], row)
                self.iBnd[n].M[iRow] = self.solver.GetVertexMach(self.bndMarkers[n], row)
                self.iBnd[n].Rho[iRow] = self.solver.GetVertexDensity(self.bndMarkers[n], row)
                if self.iBnd[n].Rho[iRow] == 0:
                    self.iBnd[n].Rho[iRow] = 1
        self.setViscousSolver(couplIter)

    def imposeBlwVel(self):
        """ Extract the solution of the viscous calculation (blowing velocity)
        and use it as a boundary condition in the inviscid solver.
        """
        if self.iBnd[0].connectListNodes[0] != self.iBnd[0].connectListNodes[-1]:
            raise RuntimeError("List of connectivity is incorrect.")

        self.getBlowingBoundary()
        # Impose blowing velocities.
        for n in range(len(self.bndMarkers)):
            if n == 0:
                self.iBnd[n].blowingVel[0] = .5 * (self.iBnd[n].blowingVel[0] + self.iBnd[n].blowingVel[-1])
                self.iBnd[n].blowingVel = self.iBnd[n].blowingVel[self.iBnd[n].connectListElems[:-1].argsort()]
            elif n == 1:
                self.iBnd[1].blowingVel = np.insert(self.iBnd[1].blowingVel, 0, self.iBnd[0].blowingVel[0])
                self.iBnd[n].blowingVel = self.iBnd[n].blowingVel[self.iBnd[n].connectListElems.argsort()]
            for iVertex in range(len(self.iBnd[n].blowingVel)-1):
                self.solver.SetBlowing(self.bndMarkers[n], iVertex, self.iBnd[n].blowingVel[iVertex])
    
    def setMesh(self):

        if self.cfg['nDim'] == 2:
            self.getAirfoil()
        elif self.cfg['nDim'] == 3:
            self.getWing()
        else:
            raise RuntimeError('su2Interface::Could not resolve how to set\
            the mesh. Either ''nDim'' is incorrect or ''msh'' was not set for\
            3D computations')

    def getAirfoil(self):
        """ Retreives the mesh used for SU2 Euler calculation.
        Airfoil is already in seilig format (Upper TE -> Lower TE).
        """
        alltags = self.solver.GetAllBoundaryMarkers()
        if 'wake' not in alltags:
            #raise RuntimeError('Can not find tag wake.')
            self.bndMarkers = [alltags['airfoil']]
            del self.iBnd[1]
            del self.vBnd[0][1]
        else:
            self.bndMarkers = [alltags['airfoil'], alltags['wake']]

        self.nPoints = [self.solver.GetNumberVertices(tag) for tag in self.bndMarkers]
        data = [np.zeros((nPts, 4)) for nPts in self.nPoints]
        for n in range(len(self.nPoints)):
            for i in range(self.nPoints[n]):
                data[n][i,:3] = self.solver.GetVertexCoord(self.bndMarkers[n], i)
                data[n][i,3] = i

        # Find TE and reorder coordinates
        teNode = data[0][:,0].argmax()
        """if teNode != 0:
            save = data[0][0,:].copy()
            data[0][0,:] = data[0][teNode,:].copy()
            data[0][teNode,:] = save
            teNode = data[0][:,0].argmax()
            print('Airfoil coordinates were reordered automatically.')"""

        for i in range(len(data[0][0,:])):
            data[0][:teNode+1,i] = np.flip(data[0][:teNode+1,i])
            data[0][teNode+1:,i] = np.flip(data[0][teNode+1:,i])

        # Duplicate TE point
        data[0] = np.row_stack((data[0], data[0][0,:]))
        # Find LE point
        self.leNode = data[0][:,0].argmin()+1
        # Recompute number of points
        self.nPoints = [len(data[n]) for n in range(len(data))]
        
        dataElems = [np.zeros((nPts-1, 3)) for nPts in self.nPoints]
        for n in range(len(data)):
            for iPoint in range(len(data[n][:,0]) - 1):
                dataElems[n][iPoint, :] = .5*(data[n][iPoint+1,:3] + data[n][iPoint, :3])
        # Nodes coordinates in the transformed space
        self.phantomNodesCoord = [d.copy() for d in data]
        self.phantomNodesCoord[0][:self.leNode, 0] *= -1

         # Compute elements cg coordinates.
        self.elemCgs = [np.zeros((nPts-1, 4)) for nPts in self.nPoints]
        for n in range(len(self.elemCgs)):
            for i in range(len(self.elemCgs[n][:,0])):
                self.elemCgs[n][i,:] = (data[n][i,:] + data[n][i+1,:]) / 2
        
        # Elements cg coordinates in the transformed space
        self.phantomElemsCoord = [ecg.copy() for ecg in self.elemCgs]
        self.phantomElemsCoord[0][:self.leNode, 0] *= -1

        for n, bnd in enumerate(self.iBnd):
            bnd.elemsCoord = dataElems[n]
            bnd.nodesCoord = data[n]
            bnd.setConnectList(data[n][:,3], data[n][:,3])
            bnd.V = np.zeros((self.nPoints[n], 3))
            bnd.M = np.zeros(self.nPoints[n])
            bnd.Rho = np.zeros(self.nPoints[n])
        self.solver.InitBlowing()
