#include "DSolver.h"
#include "DBoundaryLayer.h"
#include "DFluxes.h"
#include <Eigen/Dense>

using namespace Eigen;
using namespace tbox;
using namespace blast;

/**
 * @brief Construct a new Time Solver:: Time Solver object.
 *
 * @param _CFL0 Initial CFL number.
 * @param _Minf Freestream Mach number.
 * @param Re Freestream Reynolds number.
 * @param _maxIt Maximum number of iterations.
 * @param _tol Convergence tolerance.
 * @param _itFrozenJac Number of iterations between which the Jacobian is frozen.
 */
Solver::Solver(double _CFL0, double _Minf, double Re, unsigned int _maxIt, double _tol, unsigned int _itFrozenJac)
{
    CFL0 = _CFL0;
    maxIt = _maxIt;
    tol = _tol;
    itFrozenJac0 = _itFrozenJac;
    Minf = std::max(_Minf, 0.1);
    initSln = true;
    sysMatrix = new Fluxes(Re);
}

/**
 * @brief Destroy the Time Solver:: Time Solver object.
 *
 */
Solver::~Solver()
{
    delete sysMatrix;
    std::cout << "~blast::Solver()\n";
}

/**
 * @brief Impose initial condition at one point.
 *
 * @param iPoint Marker id.
 * @param bl Boundary layer region.
 */
void Solver::initialCondition(size_t iPoint, BoundaryLayer *bl)
{
    size_t nVar = bl->getnVar();
    if (initSln)
    {
        bl->u[iPoint * nVar + 0] = bl->u[(iPoint - 1) * nVar + 0];
        bl->u[iPoint * nVar + 1] = bl->u[(iPoint - 1) * nVar + 1];
        bl->u[iPoint * nVar + 2] = bl->u[(iPoint - 1) * nVar + 2];
        bl->u[iPoint * nVar + 3] = bl->blEdge->getVt(iPoint);
        bl->u[iPoint * nVar + 4] = bl->u[(iPoint - 1) * nVar + 4];

        bl->Ret[iPoint] = bl->Ret[iPoint - 1];
        bl->cd[iPoint] = bl->cd[iPoint - 1];
        bl->cf[iPoint] = bl->cf[iPoint - 1];
        bl->Hstar[iPoint] = bl->Hstar[iPoint - 1];
        bl->Hstar2[iPoint] = bl->Hstar2[iPoint - 1];
        bl->Hk[iPoint] = bl->Hk[iPoint - 1];
        bl->ctEq[iPoint] = bl->ctEq[iPoint - 1];
        bl->us[iPoint] = bl->us[iPoint - 1];
        bl->delta[iPoint] = bl->delta[iPoint - 1];
        bl->deltaStar[iPoint] = bl->deltaStar[iPoint - 1];
    }
    if (bl->regime[iPoint] == 1 && bl->u[iPoint * nVar + 4] <= 0)
        bl->u[iPoint * nVar + 4] = 0.03;
}

/**
 * @brief Pseudo-time integration.
 *
 * @param iPoint Marker id.
 * @param bl Boundary layer region.
 * @return int
 */
int Solver::integration(size_t iPoint, BoundaryLayer *bl)
{
    size_t nVar = bl->getnVar();

    // Save initial condition.
    std::vector<double> sln0(nVar, 0.);
    for (size_t i = 0; i < nVar; ++i)
        sln0[i] = bl->u[iPoint * nVar + i];

    // Initialize time integration variables.
    double dx = bl->mesh->getLoc(iPoint) - bl->mesh->getLoc(iPoint - 1);

    // Initial time step.
    double CFL = CFL0;
    unsigned int itFrozenJac = itFrozenJac0;
    double timeStep = setTimeStep(CFL, dx, bl->blEdge->getVt(iPoint));
    MatrixXd IMatrix(5, 5);
    IMatrix = MatrixXd::Identity(5, 5) / timeStep;

    // Initial system residuals.
    VectorXd sysRes = sysMatrix->computeResiduals(iPoint, bl);
    double normSysRes0 = sysRes.norm();
    double normSysRes = normSysRes0;

    MatrixXd JacMatrix = MatrixXd::Zero(5, 5);
    VectorXd slnIncr = VectorXd::Zero(5);

    unsigned int innerIters = 0; // Inner (non-linear) iterations

    while (innerIters < maxIt)
    {
        // Jacobian computation.
        if (innerIters % itFrozenJac == 0)
            JacMatrix = sysMatrix->computeJacobian(iPoint, bl, sysRes, 1e-8);

        // Linear system.
        slnIncr = (JacMatrix + IMatrix).partialPivLu().solve(-sysRes);

        // Solution increment.
        for (size_t k = 0; k < nVar; ++k)
            bl->u[iPoint * nVar + k] += slnIncr(k);
        bl->u[iPoint * nVar + 0] = std::max(bl->u[iPoint * nVar + 0], 1e-6);
        bl->u[iPoint * nVar + 1] = std::max(bl->u[iPoint * nVar + 1], 1.0005);

        sysRes = sysMatrix->computeResiduals(iPoint, bl);
        normSysRes = (sysRes + slnIncr / timeStep).norm();

        if (std::isnan(normSysRes))
        {
            resetSolution(iPoint, bl, sln0, true);
            sysRes = sysMatrix->computeResiduals(iPoint, bl);
            CFL = 1.0;
            timeStep = setTimeStep(CFL, dx, bl->blEdge->getVt(iPoint));
            IMatrix = MatrixXd::Identity(5, 5) / timeStep;
            normSysRes = (sysRes + slnIncr / timeStep).norm();
            itFrozenJac = 1;
        }

        if (normSysRes <= tol * normSysRes0)
            return 0; // Successfull exit.

        // CFL adaptation.
        CFL = std::max(CFL0 * pow(normSysRes0 / normSysRes, 0.7), 0.1);
        timeStep = setTimeStep(CFL, dx, bl->blEdge->getVt(iPoint));
        IMatrix = MatrixXd::Identity(5, 5) / timeStep;

        innerIters++;
    }

    if (std::isnan(normSysRes) || normSysRes / normSysRes0 > 1e-3)
    {
        resetSolution(iPoint, bl, sln0);
        return -1;
    }
    return innerIters;
}

/**
 * @brief Set time step.
 *
 * @param CFL CFL number.
 * @param dx Cell size.
 * @param invVel Inviscid velocity.
 * @return double
 */
double Solver::setTimeStep(double CFL, double dx, double invVel) const
{
    // Speed of sound.
    double gradU2 = (invVel * invVel);
    double soundSpeed = computeSoundSpeed(gradU2);

    // Velocity of the fastest travelling wave.
    double advVel = soundSpeed + invVel;

    // Time step computation.
    return CFL * dx / advVel;
}

/**
 * @brief Compute the speed of sound in a cell.
 *
 * @param gradU2 Inviscid velocity squared in the cell.
 * @return double
 */
double Solver::computeSoundSpeed(double gradU2) const
{
    return sqrt(1 / (Minf * Minf) + 0.5 * (gamma - 1) * (1 - gradU2));
}

/**
 * @brief Resets the solution of the point wrt its initial condition (sln0) or the previous point.
 *
 * @param iPoint Marker id.
 * @param bl Boundary layer region.
 * @param sln0 Initial solution at the point.
 * @param usePrevPoint if 1, use the solution at previous point instead of sln0.
 */
void Solver::resetSolution(size_t iPoint, BoundaryLayer *bl, std::vector<double> sln0, bool usePrevPoint)
{
    size_t nVar = bl->getnVar();

    // Reset solution.
    if (usePrevPoint)
        for (size_t k = 0; k < nVar; ++k)
            bl->u[iPoint * nVar + k] = bl->u[(iPoint - 1) * nVar + k];
    else
        for (size_t k = 0; k < nVar; ++k)
            bl->u[iPoint * nVar + k] = sln0[k];

    // Reset closures.
    bl->Ret[iPoint] = std::max(bl->u[iPoint * nVar + 3] * bl->u[iPoint * nVar + 0] * sysMatrix->Re, 1.0); // Reynolds number based on theta.
    bl->deltaStar[iPoint] = bl->u[iPoint * nVar + 0] * bl->u[iPoint * nVar + 1];                          // Displacement thickness.

    if (bl->regime[iPoint] == 0)
    {
        std::vector<double> lamParam(8, 0.);
        bl->closSolver->laminarClosures(lamParam, bl->u[iPoint * nVar + 0], bl->u[iPoint * nVar + 1], bl->Ret[iPoint], bl->blEdge->getMe(iPoint), bl->name);
        bl->Hstar[iPoint] = lamParam[0];
        bl->Hstar2[iPoint] = lamParam[1];
        bl->Hk[iPoint] = lamParam[2];
        bl->cf[iPoint] = lamParam[3];
        bl->cd[iPoint] = lamParam[4];
        bl->delta[iPoint] = lamParam[5];
        bl->ctEq[iPoint] = lamParam[6];
        bl->us[iPoint] = lamParam[7];
    }
    else if (bl->regime[iPoint] == 1)
    {
        std::vector<double> turbParam(8, 0.);
        bl->closSolver->turbulentClosures(turbParam, bl->u[iPoint * nVar + 0], bl->u[iPoint * nVar + 1], bl->u[iPoint * nVar + 4], bl->Ret[iPoint], bl->blEdge->getMe(iPoint), bl->name);
        bl->Hstar[iPoint] = turbParam[0];
        bl->Hstar2[iPoint] = turbParam[1];
        bl->Hk[iPoint] = turbParam[2];
        bl->cf[iPoint] = turbParam[3];
        bl->cd[iPoint] = turbParam[4];
        bl->delta[iPoint] = turbParam[5];
        bl->ctEq[iPoint] = turbParam[6];
        bl->us[iPoint] = turbParam[7];
    }
}

void Solver::setCFL0(double _CFL0)
{
    if (_CFL0 > 0)
        CFL0 = _CFL0;
    else
        throw std::runtime_error("Negative CFL numbers are not allowed.\n");
}