#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @author Paul Dechamps
# @date 2022
# Test the blast implementation. The test case is a compressible attached transitional flow.
# Tested functionalities;
# - Time integration.
# - Two time-marching techniques (agressive and safe).
# - Transition routines.
# - Forced transition.
# - Compressible flow routines.
# - Laminar and turbulent flow.
#
# Untested functionalities.
# - Separation routines.
# - Multiple failure case at one iteration.
# - Response to unconverged inviscid solver.

# Imports.

import blast.utils as viscUtils
import numpy as np
import os.path

from fwk.wutils import parseargs
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

import math

def cfgInviscid(nthrds, verb):
    # Parameters
    return {
    # Options
    'scenario' : 'aerodynamic',
    'task' : 'analysis',
    'Threads' : nthrds, # number of threads
    'Verb' : verb, # verbosity
    # Model (geometry or mesh)
    'File' : os.path.dirname(os.path.dirname(os.path.abspath(__file__))) + '/models/dart/onera_lfs.msh', # Input file containing the model
    'Pars' : {}, # parameters for input file model
    'Dim' : 3, # problem dimension
    'Format' : 'gmsh', # save format (vtk or gmsh)
    # Markers
    'Fluid' : 'field', # name of physical group containing the fluid
    'Farfield' : ['upstream', 'farfield', 'downstream'], # LIST of names of physical groups containing the farfield boundaries (upstream/downstream should be first/last element)
    'Wings' : ['wing'], # LIST of names of physical groups containing the lifting surface boundary
    'Wakes' : ['wake'], # LIST of names of physical group containing the wake
    'WakeTips' : ['wakeTip'], # LIST of names of physical group containing the edge of the wake
    'Tes' : ['te'], # LIST of names of physical group containing the trailing edge
    'dbc' : True,
    'Upstream' : 'upstream',
    # Freestream
    'M_inf' : 0.839, # freestream Mach number
    'AoA' : 3.06, # freestream angle of attack
    # Geometry
    'S_ref' : 0.7528, # reference surface length
    'c_ref' : 0.64, # reference chord length
    'x_ref' : 0.0, # reference point for moment computation (x)
    'y_ref' : 0.0, # reference point for moment computation (y)
    'z_ref' : 0.0, # reference point for moment computation (z)
    # Numerical
    'LSolver' : 'GMRES', # inner solver (Pardiso, MUMPS or GMRES)
    'G_fill' : 2, # fill-in factor for GMRES preconditioner
    'G_tol' : 1e-5, # tolerance for GMRES
    'G_restart' : 50, # restart for GMRES
    'Rel_tol' : 1e-6, # relative tolerance on solver residual
    'Abs_tol' : 1e-8, # absolute tolerance on solver residual
    'Max_it' : 50 # solver maximum number of iterations
    }

def cfgBlast(verb):
    return {
        'Re' : 1e7,       # Freestream Reynolds number
        'Minf' : 0.2,     # Freestream Mach number (used for the computation of the time step only)
        'CFL0' : 1,         # Inital CFL number of the calculation

        'sections' : np.linspace(0.026, 1.15, 20),
        'Sym':[0.],
        'span':1.196,
        'interpolator': 'Rbf',
        'rbftype': 'linear',
        'smoothing': 1e-8,
        'degree': 0,
        'neighbors': 10,

        'Verb': verb,       # Verbosity level of the solver
        'couplIter': 50,    # Maximum number of iterations
        'couplTol' : 5e-4,  # Tolerance of the VII methodology
        'iterPrint': 5,     # int, number of iterations between outputs
        'resetInv' : True,  # bool, flag to reset the inviscid calculation at every iteration.
        'xtrF' : [0., 0.],# Forced transition location
        'nDim' : 3
    }

def main():
    # Timer.
    tms = fwk.Timers()
    tms['total'].start()

    args = parseargs()
    icfg = cfgInviscid(args.k, args.verb)
    vcfg = cfgBlast(args.verb)

    parsViscous = {'nLe': 20, 'nTe': 8, 'nMid': 40, 'nSpan': 60, 'nWake': 20}
    vMsh = viscUtils.mesh(os.path.dirname(os.path.dirname(os.path.abspath(__file__))) + '/models/dart/onera_visc.geo', parsViscous)
    vcfg['vMsh'] = vMsh

    tms['pre'].start()
    coupler, iSolverAPI, vSolver = viscUtils.initBlast(icfg, vcfg)
    tms['pre'].stop()

    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    aeroCoeffs = coupler.run()
    tms['solver'].stop()

    # Display results.
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('      Re        M    alpha       Cl       Cd      Cdp      Cdf       Cm')
    print('{0:6.1f}e6 {1:8.2f} {2:8.1f} {3:8.4f} {4:8.4f} {5:8.4f} {6:8.4f} {7:8.4f}'.format(vcfg['Re']/1e6, iSolverAPI.getMinf(), iSolverAPI.getAoA()*180/math.pi, iSolverAPI.getCl(), vSolver.Cdt, vSolver.Cdp, vSolver.Cdf, iSolverAPI.getCm()))

     # Write results to file.
    vSolution = viscUtils.getSolution(vSolver)

    # Write results to file.
    for iSec in range(len(iSolverAPI.cfg['EffSections'])):
        vSolution = viscUtils.getSolution(vSolver, iSec)
        viscUtils.write(vSolution, vSolver.getRe(), sfx='slice'+str(iSec))
    vSolution['Cdt_int'] = vSolver.Cdf + iSolverAPI.getCd()

    # Save pressure coefficient
    iSolverAPI.save(sfx='_viscous')
    tms['total'].stop()

    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)
    print('SOLVERS statistics')
    print(coupler.tms)
    
    # Test solution
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('Cl', iSolverAPI.getCl(), 0.279, 5e-2))
    tests.add(CTest('Cd wake', vSolution['Cdt_int'], 0.0168, 1e-3, forceabs=True))
    tests.add(CTest('Iterations', len(aeroCoeffs), 19, 0, forceabs=True))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
